import traceback
from threading import Thread
from multiprocessing import Queue

from djangokarat.utils import (
    prepare_relational_field_to_karat, prepare_data_to_karat, send_data, get_relation_model, get_self_related_name,
    filter_relational_fields, switch_values_to_keys, append_field_name_to_keys, merge_nested_dicts,
    strip_exclamation_mark, get_karat_id_fields, strip_square_brackets, get_relation_internal_type, update_fields_in_instance
)


class Worker(Thread):
    # allow mutliple idcolumns to retrieve 1 object
    # drop support for non-dict karat_fields
    queue = Queue()
    models = None

    def run(self):
        while True:
            method, args = self.queue.get()
            try:
                method(*args)
            except Exception:
                print(traceback.format_exc())

    @classmethod
    def add(cls, method, *args):
        cls.assign_karat_models()
        cls.queue.put((method, *args))

    @classmethod
    def add_sync(cls, data_array):
        cls.assign_karat_models()
        cls.add(cls.sync, [cls, data_array])

    @classmethod
    def assign_karat_models(cls):
        if cls.models:
            return
        # sets all models with karat attributes to local variable for later use
        from django.apps import apps
        models = apps.get_models()
        cls.models = list(((model._meta.original_attrs['karat_table'], model)
                           for model in models if 'karat_table' in model._meta.original_attrs))

    def sync(self, data_array):
        self.assign_karat_models()
        # all incoming data are parsed by their table names
        for data in data_array:
            for key in data.keys():
                used_tables = self.find_index_of_tables(self, key)
                self.add(self.update_or_create_instance, [self, used_tables, data[key]])

    def find_index_of_tables(self, table):
        # based on incoming data gets all indexes of models with `karat_table` where data will be saved
        tables_index = [index for index, model in enumerate(self.models) if model[0] == table]
        return tables_index

    def skip_empty_model(self, data, karat_fields):
        '''
        if there are multiple models in one table in karat, check if ours model has to be modified
        '''
        if isinstance(karat_fields, dict):
            should_skip = False
            id_columns = get_karat_id_fields(karat_fields)
            model_fields = switch_values_to_keys(karat_fields)
            for id_column in id_columns:
                # if there are data for field and field shouldn't be ignored
                if not str(data.get(id_column, '')).strip():
                    should_skip = True
        return should_skip

    def prepare_data_to_send(self, used_instances):
        ''' cycle to used instances and build them to single dictionary
        '''
        data_to_send = {}
        for instance in used_instances:
            data_to_send = {**prepare_data_to_karat(instance), **data_to_send}
            data_to_send['_table'] = instance._meta.karat_table
        return data_to_send

    def build_instance_map(self, model, fields, karat_fields, data):
        instance_map = {}
        for field in fields:
            parts = field.split('__')
            data_name = karat_fields[field]
            lookup_model_dict = self.build_relational_map_lookup(self, model, parts, data[data_name])
            instance_map = merge_nested_dicts(instance_map, lookup_model_dict)
        return instance_map

    def build_relational_map_lookup(self, model, parts, data_value):
        field_name = parts.pop(0)
        lookup_dictionary = {}
        if parts:
            field_model = get_relation_model(model, field_name)
            lookup_dictionary[field_name] = self.build_relational_map_lookup(self, field_model, parts, data_value)
            return lookup_dictionary
        else:
            return {field_name: data_value}

    def build_instance_from_map(self, model, instance_map):
        ''' gets instance based of input dictionary of values
        '''
        models_instance_map, models_instance_id_map = self.build_models_in_instance_map(self, model, instance_map)
        instance = self.get_or_create_instance_from_dict(self, model, models_instance_map)
        return instance, models_instance_id_map

    def update_instance_data_from_map(self, instance, instance_map, instance_id_dict):
        updated_instance = self.update_or_create_instance_from_dict(self, instance, instance_map, instance_id_dict)
        return updated_instance

    def build_models_in_instance_map(self, model, instance_map, level=0):
        ''' recursively goes through built dictionary and each dict objects turns into id of object
        returns tuple.
        '''
        instance_id_only_map = {}
        has_dict = True
        while has_dict:
            # premise that dictionary hasn't dictionary
            has_dict = False
            # upfront check if in instance_map is any dictionary
            for instance_model_name, data in instance_map.items():
                if isinstance(data, dict):
                    has_dict = True

            for instance_model_name, data in instance_map.items():
                if isinstance(data, dict):
                    instance_model = get_relation_model(model, instance_model_name)
                    instance_model_relation_type = get_relation_internal_type(model, instance_model_name)
                    # recurse to get instance model instance
                    instance, instance_id_map = self.build_models_in_instance_map(
                        self, instance_model, data, level + 1)
                    instance_id = instance.pk
                    # after id is found, field name has to have `_id` appended
                    instance_model_name_with_id = "{}_id".format(instance_model_name)

                    instance_id_only_map[instance_model_name] = instance_id_map
                    # add and `_id` field name
                    if instance_model_relation_type == 'OneToOneField':
                        instance_map[instance_model_name] = instance
                    else:
                        instance_map[instance_model_name_with_id] = instance_id
                        # remove original name
                        instance_map.pop(instance_model_name)

                elif not has_dict and level > 0:
                    instance = self.get_or_create_instance_from_dict(self, model, instance_map)
                    instance_id_only_map['id'] = instance.id
                    return instance, instance_id_only_map
                else:
                    continue

        return instance_map, instance_id_only_map

    def get_or_create_instance_from_dict(self, model, data_dict):
        instance_model_filter = model.objects.filter(**data_dict)

        if instance_model_filter.exists():
            instance = instance_model_filter.get()
            return instance
        else:
            instance = model(**data_dict)
            if instance._meta.karat_fields:
                instance.save(_send=False)
            else:
                instance.save()
            return instance

    def update_or_create_instance_from_dict(self, instance, data_dict, instance_id_dict):
        for name, value in data_dict.items():
            if isinstance(value, dict):
                new_object = False
                inner_instance = getattr(instance, name, None)
                if name not in instance_id_dict or 'id' not in instance_id_dict[name]:
                    new_object = True
                inner_instance_model = get_relation_model(instance._meta.model, name)
                if not inner_instance and not new_object:
                    inner_instance = inner_instance_model.objects.get(id=instance_id_dict[name]['id'])

                if new_object:
                    inner_instance = self.get_or_create_instance_from_dict(self, inner_instance_model, value)
                    setattr(instance, name, inner_instance)
                else:
                    self.update_or_create_instance_from_dict(self, inner_instance, value, instance_id_dict[name])
            else:
                setattr(instance, name, value)
        if instance._meta.karat_fields:
            instance.save(_send=False)
        else:
            instance.save()
        return instance

    def has_map_ready(self, instance_relation_map):
        for field, value in instance_relation_map.items():
            if isinstance(value, dict):
                return self.has_map_ready(self, value)
            else:
                if value is None:
                    return False
                elif isinstance(value, str) and not value.strip():
                    return False
                else:
                    return True

    def update_or_create_instance(self, tables, data):
        used_instances = []
        # incoming data are sorted to their according tables by its unique id
        for table in tables:
            name, model = self.models[table]
            karat_fields = model._meta.karat_fields.copy()
            id_columns = get_karat_id_fields(karat_fields)
            # remove model id column so it cannot be changed in db
            check_model_fields = karat_fields.copy()
            # remove syncing column since it should be in every model (specified in constraints)
            for id_column in id_columns:
                check_model_fields.pop(id_column, None)

            # if model doesn't have fields to be updated skip it
            if self.skip_empty_model(self, data, check_model_fields):
                continue

            if isinstance(karat_fields, dict):
                # flip values for keys
                model_fields = switch_values_to_keys(karat_fields)
                # strip field_names of exclamation marks and brackets
                for field in model_fields.keys():
                    model_fields[field] = strip_exclamation_mark(model_fields[field])
                    model_fields[field] = strip_square_brackets(model_fields[field])
                # extract lookups from model
                lookup_fields = filter_relational_fields(model_fields.values())
                karat_fields = switch_values_to_keys(model_fields)
            else:
                model_fields = dict((field, field) for field in karat_fields)

            id_fields = [model_fields[id_column] for id_column in id_columns]
            data_fields = list(set(model_fields.values()) - set(id_fields))

            instance_relation_map = self.build_instance_map(self, model, id_fields, karat_fields, data)
            if not self.has_map_ready(self, instance_relation_map):
                continue
            instance_data_map = self.build_instance_map(self, model, data_fields, karat_fields, data)
            instance, instance_id_dict = self.build_instance_from_map(self, model, instance_relation_map)
            updated_instance = self.update_instance_data_from_map(self, instance, instance_data_map, instance_id_dict)
            used_instances.append(updated_instance)

        # data_to_send = self.prepare_data_to_send(self, used_instances)
        # self.add(send_data, [data_to_send])

worker = Worker()
worker.daemon = True
worker.start()
name = "djangokarat"



import requests
import json

from django.conf import settings
from django.forms.models import model_to_dict
from django.core.serializers.json import DjangoJSONEncoder

# TODO remove zfill in data send


def has_karat_fields(instance):
    if not instance._meta.karat_table and not instance._meta.karat_fields:
        return False
    return True


def convert__all__to_fields(instance):
    # Checks meta karat_fields for __all__ to add to fields
    if (instance._meta.karat_table and not instance._meta.karat_fields) or instance._meta.karat_fields == '__all__':
        instance._meta.karat_fields = [f.name for f in instance._meta.get_fields()]


def get_karat_id_fields(karat_fields):
    fields = []
    if isinstance(karat_fields, dict):
        for field in karat_fields.keys():
            if check_start_end_square_brackets(field):
                fields.append(karat_fields[field])

    else:
        for field in karat_fields:
            if check_start_end_square_brackets(field):
                fields.append(field)

    if not fields:
        raise AttributeError
    return fields


def filter_relational_fields(list):
    return [field for field in list if '__' in field]


def strip_exclamation_mark(field):
    if '!' not in field:
        return field
    field = field.replace('!', '')
    return field


def switch_values_to_keys(dict_to_swap):
    if not isinstance(dict_to_swap, dict):
        raise TypeError
    return dict((v, k) for k, v in dict_to_swap.items())


def check_start_end_square_brackets(field):
    if not field.startswith('[') or not field.endswith(']'):
        return False
    return True


def strip_square_brackets(field):
    if '[' not in field and ']' not in field:
        return field

    square_brackets = '[]'
    for bracket in square_brackets:
        field = field.replace(bracket, '')
    return field


def get_relation_model(model, field_name):
    # get model of relation field (FK, O2O) >= Django 2.0
    return model._meta.get_field(field_name).remote_field.model


def get_self_related_name(model, field_name):
    # get how attribute is named in referencing model
    return model._meta.get_field(field_name).remote_field.related_name


def get_relation_internal_type(model, field_name):
    # returns string value of relation (foreignkey, onetoonefield)
    return model._meta.get_field(field_name).get_internal_type()


def append_field_name_to_keys(dict_to_change, field_name):
    switched_dict = switch_values_to_keys(dict_to_change)
    for key, value in switched_dict.items():
        switched_dict[key] = "{}__{}".format(field_name, value)
    return switch_values_to_keys(switched_dict)


def update_fields_in_instance(instance, data_dict):
    for attribute, value in data_dict.items():
        setattr(instance, attribute, value)
    instance.save()
    return instance


def prepare_relational_field_to_karat(instance, field):
    ''' get data of instance at the end of relational field  
    '''
    fields = field.split('__')
    field_data = instance
    for field in fields:
        if field == 'id':
            field_data = getattr(field_data, field, None)
            field_data = str(field_data).zfill(8)
        else:
            field_data = getattr(field_data, field, None)
    return field_data


def merge_nested_dicts(base_dict, merging_dict, path=None):
    ''' immutates base dict and extends it with merging dict
    '''
    if path is None:
        path = []
    for key in merging_dict:
        if key in base_dict:
            if isinstance(base_dict[key], dict) and isinstance(merging_dict[key], dict):
                merge_nested_dicts(base_dict[key],  merging_dict[key], path + [str(key)])
            elif base_dict[key] == merging_dict[key]:
                pass  # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            base_dict[key] = merging_dict[key]
    return base_dict


def prepare_data_to_karat(instance):
    switched_karat_fields = switch_values_to_keys(instance._meta.karat_fields)
    cleaned_karat_fields = switched_karat_fields.copy()
    # clean fields of controlling tokens
    for name, value in switched_karat_fields.items():
        cleaned_karat_fields[name] = strip_square_brackets(cleaned_karat_fields[name])
        if '!' in value:
            cleaned_karat_fields.pop(name)
    switched_karat_fields = cleaned_karat_fields
    # have dictionary ready to get karat field names
    karat_fields = switch_values_to_keys(switched_karat_fields)
    # gets all flat data (non-relational)
    model_data = model_to_dict(instance, fields=switched_karat_fields.values())
    # build dictionary with keys from karat db and with flat data
    karat_data = {name: model_data[value] for name, value in switched_karat_fields.items() if value in model_data}
    relational_fields = filter_relational_fields(switched_karat_fields.values())
    # get data of relational fields
    for relation_field in relational_fields:
        field = relation_field.split('__')[0]
        new_model = getattr(instance, field)
        field_name = karat_fields[relation_field]
        if not new_model:
            continue
        if new_model._meta.karat_table == instance._meta.karat_table:
            karat_data = {**prepare_data_to_karat(new_model), **karat_data}
        else:
            karat_data[field_name] = prepare_relational_field_to_karat(instance, relation_field)
    return karat_data


def send_data(data):
    print(data)
    if hasattr(settings, 'AGENT_URL'):
        r = requests.post('{}/accept-data/'.format(settings.AGENT_URL),
                          data=json.dumps(data, cls=DjangoJSONEncoder))



from django.urls import path, include

from djangokarat.views import SyncData, GetKarat

urlpatterns = [
    path('karat/sync', SyncData.as_view(), name='karat-sync'),
    path('karat/ping', GetKarat.as_view(), name='karat-ping'),
]



import requests
from rest_framework.views import APIView
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from django.http import JsonResponse
from django.views.generic import View
from django.conf import settings


import json
from . import Worker


class SyncData(APIView):
    permission_classes = (AllowAny,)

    def post(self, request):
        Worker.add_sync(json.loads(request.data))
        return Response({}, status=200)


class GetKarat(View):

    def get(self, request):
        if not hasattr(settings, 'AGENT_URL'):
            return JsonResponse({"message": "missing AGENT_URL"}, status=409)
        url = "{}{}".format(settings.AGENT_URL, '/ping-sync/')
        r = requests.get(url)
        return JsonResponse(r.text, status=200, safe=False)



from django.db.models import Model

from djangokarat.constraints import CheckKaratFields
from djangokarat.utils import (
    has_karat_fields, convert__all__to_fields, strip_square_brackets, prepare_data_to_karat,
    get_karat_id_fields, filter_relational_fields, switch_values_to_keys, send_data,
)

from django.db import models
models.options.DEFAULT_NAMES += ('karat_table', 'karat_fields')
models.options.Options.karat_fields = None
models.options.Options.karat_table = None

from . import Worker


class KaratModel(Model):
    # create delete call
    # check thru multiple tables same unique name
    # change how exclamation mark works

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        convert__all__to_fields(self)

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        convert__all__to_fields(cls)
        errors.extend(cls._check_karat_meta(**kwargs))
        return errors

    @classmethod
    def _check_karat_meta(cls, **kwargs):
        errors = []
        # if karat params aren't there, don't check errors for them
        if not has_karat_fields(cls):
            return errors

        karat_check = CheckKaratFields(cls)
        # upfront model checks
        errors.extend(karat_check.check_karat_table())
        errors.extend(karat_check.check_exclamation_fields())
        errors.extend(karat_check.check_karat_fields())
        errors.extend(karat_check.check_uniqueness())
        errors.extend(karat_check.check_lookups())
        return errors

    def save(self, _send=True, enable_exists_check=False, extra_kwargs={}, *args, **kwargs):
        data_kwargs = {}
        if enable_exists_check:
            exists = self._meta.model.objects.filter(id=self.id).exists()
            data_kwargs = {'enable_exists_check': enable_exists_check, 'update': exists, **extra_kwargs}

        super().save(*args, **kwargs)
        if not has_karat_fields(self):
            return
        if _send:
            model_data = prepare_data_to_karat(self)
            prepared_data = self.prepare_data(model_data, **data_kwargs)
            Worker.add(send_data, [prepared_data])

    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        # if not has_karat_fields(self):
        #     return
        # unique_field = get_karat_id_fields(self)

    def prepare_data(self, karat_fields_dict, *args, **kwargs):
        karat_fields_dict['_table'] = self._meta.karat_table
        karat_fields_dict['_description'] = str(self)
        return karat_fields_dict



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="djangokarat",
    version="1.2.2",
    author="David Simonyi",
    author_email="david@whys.eu",
    description="ability to create models with send/accept fields to/from karat DB",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='Creative Commons Attribution-Noncommercial-Share Alike license',
    url="https://kiwi.whys.eu/karaterp/axima-agent",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



from django.db.models.sql.query import Query
from django.core import checks
from djangokarat.utils import (filter_relational_fields, switch_values_to_keys, strip_square_brackets,
                               strip_exclamation_mark, check_start_end_square_brackets)


class CheckKaratFields:

    def __init__(self, cls):
        self.cls = cls
        self.karat_table = cls._meta.karat_table
        self.karat_fields = cls._meta.karat_fields
        self.model_fields = cls._meta.get_fields()
        self.model_fields_names = [field.name for field in self.model_fields]
        self.is_dict = isinstance(self.karat_fields, dict)

    def is_relation(self, field):
        if '__' in field:
            return True
        return False

    def extract_field_name_from_lookups(self, field_array=None):
        ''' gets all fields with `__`. Extracts first part of name and replaces 
        lookup with only field name
        '''
        if not field_array:
            if self.is_dict:
                field_array = self.karat_fields.keys()
            else:
                field_array = self.karat_fields
        field_lookups = [field for field in field_array if '__' in field]
        field_array = list(set(field_array) - set(field_lookups))
        lookups_first_part = []
        for lookup in field_lookups:
            if check_start_end_square_brackets(lookup):
                stripped_lookup = strip_square_brackets(lookup)
                lookups_first_part.append('[{}]'.format(stripped_lookup.split('__')[0]))
            else:
                lookups_first_part.append(lookup.split('__')[0])

        return field_array + lookups_first_part

    def strip_miscellaneous_chars_from_fields(self, strip_function):
        ''' remove all unwanted characters from karat_fields to work with only field 
        names without i.e. `!`
        '''
        if self.is_dict:
            switched_karat_fields = switch_values_to_keys(self.karat_fields)
            for key in switched_karat_fields.keys():
                switched_karat_fields[key] = strip_function(switched_karat_fields[key])
            self.karat_fields = switch_values_to_keys(switched_karat_fields)
        else:
            for index, field in enumerate(self.karat_fields):
                self.karat_fields[index] = strip_function(field)

    def check_karat_table(self):
        errors = []
        if not self.karat_table and self.karat_fields:
            errors.append(
                checks.Error(
                    'missing field `karat_table` in Meta',
                    hint="add field `karat_table` to Meta or remove field `karat_fields`",
                    obj=self.cls,
                    id='karat.E001'
                )
            )
        elif not isinstance(self.karat_table, str):
            errors.append(
                checks.Error(
                    'field `karat_table` must be string',
                    obj=self.cls,
                    id='karat.E002'
                )
            )
        if self.karat_table and not self.karat_fields:
            errors.append(
                checks.Warning(
                    'missing field `karat_fields` in Meta',
                    hint="if not added all fields be used to karat_fields similiar to `__all__`",
                    obj=self.cls,
                    id='karat.W001'
                )
            )
        return errors

    def check_karat_fields(self):
        errors = []
        if (
            (isinstance(self.karat_fields, str) and self.karat_fields != '__all__')
            or not isinstance(self.karat_fields, (str, dict, list))
        ):
            errors.append(
                checks.Error(
                    'field `karat_fields` has to be `__all__`, list or dict',
                    hint="change field to correct type",
                    obj=self.cls,
                    id='karat.E003'
                )
            )
        if self.karat_fields and not '__all__':
            for field in self.karat_fields:
                if not field in self.model_fields_names:
                    errors.append(
                        checks.Error(
                            "doesn't have field {}".format(field),
                            hint="Correct typo in model",
                            obj=self.cls,
                            id='karat.E004'
                        )
                    )
        elif self.is_dict:
            karat_fields = self.extract_field_name_from_lookups(self.karat_fields.keys())
            for field in karat_fields:
                stripped_field = strip_square_brackets(field)
                if stripped_field not in self.model_fields_names:
                    errors.append(
                        checks.Error(
                            "doesn't have field {}".format(field),
                            hint="Correct typo in model key",
                            obj=self.cls,
                            id='karat.E004'
                        )
                    )

        return errors

    def check_uniqueness(self):
        errors = []
        unique = False

        if self.is_dict:
            karat_fields_keys = self.karat_fields.keys()
            karat_fields = self.extract_field_name_from_lookups(karat_fields_keys)
        else:
            karat_fields = self.karat_fields

        for field in karat_fields:
            if check_start_end_square_brackets(field):
                unique = True
                stripped_field = strip_square_brackets(field)
                model_field = self.cls._meta.get_field(stripped_field)
                model_field_internal_type = model_field.get_internal_type()
                if model_field_internal_type != 'OneToOneField' and model_field_internal_type != 'ForeignKey' and not model_field.unique:
                    errors.append(
                        checks.Warning(
                            "field is used as pk for connecting databases, but isn't unique",
                            hint="change {} to unique".format(field),
                            obj=self.cls,
                            id='karat.W002'
                        )
                    )

        if not unique:
            errors.append(
                checks.Error(
                    "class doesn't have unique in `karat_fields` used to connect dbs",
                    hint="Add unique field for id from karat db",
                    obj=self.cls,
                    id='karat.E005'
                )
            )
        self.strip_miscellaneous_chars_from_fields(strip_square_brackets)
        return errors

    def check_lookups(self):
        if self.is_dict:
            relational_fields = filter_relational_fields(self.karat_fields.keys())
        else:
            relational_fields = filter_relational_fields(self.karat_fields)

        errors = []
        if not relational_fields:
            return errors

        query = Query(self.cls)
        for field in relational_fields:
            lookups, parts, _ = query.solve_lookup_type(field)
            if lookups:
                errors.append(
                    checks.Error(
                        "There can't be lookups or not existing fields in relation",
                        hint="Fix {} to point to model field".format(field),
                        obj=self.cls,
                        id='karat.E006'
                    )
                )
        return errors

    def check_exclamation_error(self, field):
        if not field.startswith('!') and '!' in field:
            return checks.Error(
                "Exclamation mark must be at the beginning of field",
                hint="Remove or move exclamation mark in {}".format(field),
                obj=self.cls,
                id='karat.E007'
            )

    def check_exclamation_fields(self):
        errors = []

        if self.is_dict:
            for key in self.karat_fields.keys():
                if '!' not in key:
                    continue
                error = self.check_exclamation_error(key)
                if error:
                    errors.append(error)
        else:
            for field in self.karat_fields:
                if '!' not in field:
                    continue
                error = self.check_exclamation_error(field)
                if error:
                    errors.append(error)
        self.strip_miscellaneous_chars_from_fields(strip_exclamation_mark)
        return errors
